import cv_depot.core.enum  # noqa F401
import cv_depot.core.channel_map  # noqa F401
import cv_depot.core.color  # noqa F401
import cv_depot.core.enforce  # noqa F401
import cv_depot.core.image  # noqa F401
import cv_depot.core.tools  # noqa F401
import cv_depot.core.types  # noqa F401
import cv_depot.core.video  # noqa F401
import cv_depot.core.viewer  # noqa F401



from typing import Any  # noqa F401
from cv_depot.core.types import AnyAnchor, AnyColor, OptArray  # noqa F401

import math

from lunchbox.enforce import Enforce
import cv2
import numpy as np

from cv_depot.core.color import BasicColor
from cv_depot.core.enum import Anchor
from cv_depot.core.image import BitDepth, Image
import cv_depot.core.enforce as enf
import cv_depot.ops.draw as cvdraw
# ------------------------------------------------------------------------------


def reformat(image, width, height):
    # type: (Image, float, float) -> Image
    '''
    Reformat given image by given width and height factors.

    Args:
        image (Image): Image instance.
        width (float): Factor to scale image width by.
        height (float): Factor to scale image height by.

    Raises:
        EnforceError: If image is not an Image instance.
        ValueError: If width or height of reformatted image is less than 1 pixel.

    Returns:
        Image: Reformatted image.
    '''
    Enforce(image, 'instance of', Image)
    # --------------------------------------------------------------------------

    source_bit_depth = image.bit_depth
    image = image.to_bit_depth(BitDepth.FLOAT32)

    x = round(image.width * width)
    y = round(image.height * height)
    if x < 1 or y < 1:
        msg = 'Invalid scale factors. Width and height must be at least 1 pixel'
        msg += f'. Format shape in pixels: ({width}, {height}).'
        raise ValueError(msg)

    tmp = cv2.resize(image.data, (x, y))
    output = Image.from_array(tmp).to_bit_depth(source_bit_depth)
    return output


def crop(image, width_mult, height_mult, width_offset=0, height_offset=0):
    # type: (Image, float, float, int, int) -> Image
    '''
    Crop a given image according to a width and height multipliers and offsets.

    Args:
        image (Image): Image to be cropped.
        width_mult (float): Width multiplier.
        height_mult (float): Height multiplier.
        width_offset (float): Width offset.
        height_offset (float): Height offset.

    Raises:
        EnforeError: If image is not an instance of Image.
        EnforeError: If width_mult is not > 0 and <= 1.
        EnforeError: If height_mult is not > 0 and <= 1.
        EnforceError: If crop dimensions are 0 in width or height.
        EnforceError: If crop width bounds are outside image dimensions.
        EnforceError: If crop height bounds are outside image dimensions.

    Returns:
        Image: Cropped image.
    '''
    Enforce(image, 'instance of', Image)
    Enforce(width_mult, '>', 0)
    Enforce(width_mult, '<=', 1)
    Enforce(height_mult, '>', 0)
    Enforce(height_mult, '<=', 1)
    # --------------------------------------------------------------------------

    height_offset *= -1

    w, h = image.width_and_height
    cw = w * 0.5
    ch = h * 0.5
    wb = cw * width_mult
    hb = ch * height_mult

    w0 = int((cw - wb) + width_offset)
    w1 = int(cw + wb + width_offset)
    h0 = int((ch - hb) + height_offset)
    h1 = int(ch + hb + height_offset)

    cw = int(cw)
    ch = int(ch)

    # ensure crop bounds are within image dimensions
    dw = int(w * width_mult)
    dh = int(h * height_mult)
    msg = 'Crop width and height must be greater than 0. '
    msg += f'Crop dimensions: {dw}, {dh}'
    Enforce(w * width_mult, '>=', 1, message=msg)
    Enforce(h * height_mult, '>=', 1, message=msg)

    msg = f'Crop width bounds: ({w0 - cw}, {w1 - cw}) outside of '
    msg += f'image width bounds: ({-cw}, {cw})'
    Enforce(w0, '>=', 0, message=msg)
    Enforce(w1, '>=', 0, message=msg)
    Enforce(w0, '<=', w, message=msg)
    Enforce(w1, '<=', w, message=msg)

    msg = f'Crop height bounds: ({h0 - ch}, {h1 - ch}) outside of '
    msg += f'image height bounds: ({-ch}, {ch})'
    Enforce(h0, '>=', 0, message=msg)
    Enforce(h1, '>=', 0, message=msg)
    Enforce(h0, '<=', h, message=msg)
    Enforce(h1, '<=', h, message=msg)

    return image[w0:w1, h0:h1]


def pad(image, shape, anchor=Anchor.TOP_LEFT, color=BasicColor.BLACK.name):
    # type: (Image, tuple[int, int, int], AnyAnchor, AnyColor) -> Image
    '''
    Pads a given image into a new image of a given shape.

    The anchor argument determines which corner of the given image will be
    anchored to the padded images respective corner. For instance, an anchor of
    'top-left' will anchor the top-left corner of the given image to the
    top-left corner of the padded image. 'center-left' will vertically center
    the image and horizontally anchor to the left of the image. 'center-center'
    will vertically and horizontally center the image.

    Args:
        image (Image): Image to be padded.
        shape (tuple[int]): Shape (width, height, channels) of padded image.
        anchor (Anchor or str, optional): Where the given image will be placed within the
            new image. Default: top-left.
        color (Color or BasicColor, optional): Padding color.
            Default: BasicColor.BLACK

    Returns:
        Image: Padded image.
    '''
    if len(shape) != 3:
        msg = f'Shape must be of length 3. Given shape: {shape}.'
        raise ValueError(msg)

    w, h, c = image.shape
    dw = shape[0] - w
    dh = shape[1] - h
    c = max(c, shape[2])

    if dw < 0 or dh < 0:
        msg = 'Output shape must be greater than or equal to input shape in each'
        msg += f' dimension. {shape} !>= {image.shape}.'
        raise ValueError(msg)
    # --------------------------------------------------------------------------

    # resolve anchor
    if isinstance(anchor, str):
        dir_h, dir_w = Anchor.from_string(anchor).value
    else:
        dir_h, dir_w = anchor.value

    lut = dict(
        center='center', top='below', bottom='above', left='right', right='left'
    )
    dir_h = lut[dir_h]
    dir_w = lut[dir_w]
    # --------------------------------------------------------------------------

    bit_depth = image.bit_depth
    output = image.to_bit_depth(BitDepth.FLOAT32)
    if dh > 0:
        if dir_h != 'center':
            pad_h = cvdraw.swatch((output.width, dh, c), color)
            output = staple(output, pad_h, direction=dir_h)

        else:
            ph0 = (output.width, math.ceil(dh / 2), c)
            ph1 = (output.width, math.floor(dh / 2), c)

            pad_h0 = cvdraw.swatch(ph0, color)
            output = staple(output, pad_h0, direction='above')

            if ph1[1] > 0:
                pad_h1 = cvdraw.swatch(ph1, color)
                output = staple(output, pad_h1, direction='below')

    if dw > 0:
        if dir_w != 'center':
            pad_w = cvdraw.swatch((dw, output.height, c), color)
            output = staple(output, pad_w, direction=dir_w)

        else:
            pw0 = (math.ceil(dw / 2), output.height, c)
            pw1 = (math.floor(dw / 2), output.height, c)

            pad_w0 = cvdraw.swatch(pw0, color)
            output = staple(output, pad_w0, direction='left')

            if pw1[0] > 0:
                pad_w1 = cvdraw.swatch(pw1, color)
                output = staple(output, pad_w1, direction='right')

    output = output.to_bit_depth(bit_depth)
    return output


def staple(image_a, image_b, direction='right', fill_value=0.0):
    # type: (Image, Image, str, float) -> Image
    '''
    Joins two images along a given direction.

    .. image:: images/staple.png

    Images must be the same height if stapling along left/right axis.
    Images must be the same width if stapling along above/below axis.

    Args:
        image_a (Image): Image A.
        image_b (Image): Image B.
        direction (str, optional): Where image b will be placed relative to a.
            Options include: left, right, above, below. Default: right.
        fill_value (float, optional): Value to fill additional channels with.
            Default: 0.

    Raises:
        ValueError: If illegal direction given.
        ValueError: If direction is left/right and image heights are not
            equal.
        ValueError: If direction is above/below and image widths are not
            equal.

    Returns:
        Image: Stapled Image.
    '''
    direction = direction.lower()
    dirs = ['left', 'right', 'above', 'below']
    if direction not in dirs:
        msg = f'Illegal direction: {direction}. Legal directions: {dirs}.'
        raise ValueError(msg)

    if direction in ['left', 'right']:
        h0 = image_a.height
        h1 = image_b.height
        if h0 != h1:
            msg = f'Image heights must be equal. {h0} != {h1}.'
            raise ValueError(msg)

    elif direction in ['above', 'below']:
        w0 = image_a.width
        w1 = image_b.width
        if w0 != w1:
            msg = f'Image widths must be equal. {w0} != {w1}.'
            raise ValueError(msg)

    # pad images so number of channels are equal
    a = image_a.data
    b = image_b.data

    # needed for one channel images which make (h, w) arrays
    if len(a.shape) < 3:
        a = np.expand_dims(a, axis=2)
    if len(b.shape) < 3:
        b = np.expand_dims(b, axis=2)

    ca = image_a.num_channels
    cb = image_b.num_channels
    if ca != cb:
        w, h, _ = image_a.shape
        c = abs(ca - cb)
        pad = np.ones((h, w, c), dtype=np.float32) * fill_value
        if ca > cb:
            b = np.concatenate([b, pad], axis=2)
        else:
            a = np.concatenate([a, pad], axis=2)

    if direction == 'above':
        data = np.append(b, a, axis=0)
    elif direction == 'below':
        data = np.append(a, b, axis=0)
    elif direction == 'left':
        data = np.append(b, a, axis=1)
    elif direction == 'right':
        data = np.append(a, b, axis=1)

    return Image.from_array(data)


def cut(image, indices, axis='vertical'):
    # (Image, Union[int, list[int]], str) -> Image
    '''
    Splits a given image into two images along a vertical or horizontal axis.

    .. image:: images/cut.png

    Args:
        image (Image): Image to be cut.
        indices (int or list[int]): Indices of where to cut along cross-axis.
        axis (str, optional): Axis to cut along.
            Options include: vertical, horizontal. Default: vertical.

    Raises:
        EnforceError: If image is not an Image instance.
        EnforceError: If indices is not an int or list of ints.
        EnforceError: If illegal axis is given.
        IndexError: If indices contains index that is outside of bounds.

    Returns:
        tuple[Image]: Two Image instances.
    '''
    Enforce(image, 'instance of', Image)

    axis = axis.lower()
    msg = 'Illegal axis: {a}. Legal axes include: {b}.'
    Enforce(axis, 'in', ['vertical', 'horizontal'], message=msg)

    # create indices
    if isinstance(indices, int):
        indices = [indices]

    Enforce(indices, 'instance of', list)
    enf.enforce_homogenous_type(indices)
    Enforce(indices[0], 'instance of', int)  # tyep: ignore

    indices.append(0)
    max_ = image.width
    if axis == 'horizontal':
        max_ = image.height
    indices.append(max_)
    indices = sorted(list(set(indices)))

    max_i = max(indices)
    if max_i > max_:
        msg = f'Index out of bounds. {max_i} > {max_}.'
        raise IndexError(msg)

    min_i = min(indices)
    if min_i < 0:
        msg = f'Index out of bounds. {min_i} < 0.'
        raise IndexError(msg)
    # --------------------------------------------------------------------------

    output = []
    for j, i in enumerate(indices):
        if j == 0:
            continue
        prev = indices[j - 1]
        if axis == 'vertical':
            img = image[prev:i, :, :]
        else:
            img = image[:, prev:i, :]
        output.append(img)
    return output


def chop(image, channel='a', mode='vertical-horizontal'):
    # type: (Image, str, str) -> dict[tuple[int, int], Image]
    '''
    Chops up a given image into smaller images that bound single contiguous
    objects within a given channel.

    .. image:: images/chop_example.png

    Chop has the following modes:

    .. image:: images/chop_modes.png

    Args:
        image (Image): Image instance.
        channel (str, optional): Channel to chop image by. Default: 'a'.
        mode (str, optional): The type and order of cuts to ber performed.
            Default: vertical-horizontal. Options include:

                * vertical - Make only vertical cuts along the width axis.
                * horizontal - Make only horizontal cuts along the height axis.
                * vertical-horizontal - Cut along the width axis first and then
                    the height axis of each resulting segement.
                * horizontal-vertical - Cut along the height axis first and then
                    the width axis of each resulting segement.

    Raises:
        EnforceError: If image is not an Image or NDArray.
        EnforceError: If channel is not in image channels.
        EnforceError: If illegal mode given.

    Returns:
        dict: Dictionary of form (width, height): Image.
    '''
    # enforce image
    if isinstance(image, np.ndarray):
        image = Image.from_array(image)
    Enforce(image, 'instance of', Image)

    # enforce channel
    msg = '{a} is not a valid channel. Channels include: {b}.'
    Enforce(channel, 'in', image.channels, message=msg)

    # enforce mode
    modes = [
        'vertical', 'horizontal', 'vertical-horizontal', 'horizontal-vertical'
    ]
    msg = '{a} is not a legal mode. Legal modes include: {b}.'
    Enforce(mode, 'in', modes, message=msg)
    # --------------------------------------------------------------------------

    def get_bounding_boxes(image):
        # type: (Image) -> list[list[tuple[int, int]]]
        hgt = image.height
        img = image.to_bit_depth(BitDepth.UINT8).data
        contours = cv2.findContours(
            img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )
        items = contours[0] if len(contours) == 2 else contours[1]

        output = []
        for item in items:
            x, y, w, h = cv2.boundingRect(item)
            output.append([
                (x, hgt - (y + h)),
                (x + w, hgt - y),
            ])
        return output

    def get_indices(image, axis):
        # type: (Image, str) -> list[int]
        # find indices of bbox edges across a single axis
        bboxes = get_bounding_boxes(image)
        indices = set()  # type: Any
        if axis == 'vertical':
            for bbox in bboxes:
                indices.add(bbox[0][0])
                indices.add(bbox[1][0])
        else:
            for bbox in bboxes:
                indices.add(bbox[0][1])
                indices.add(bbox[1][1])
        indices = list(indices)

        max_ = image.width if axis == 'vertical' else image.height
        indices = filter(lambda x: x not in [0, max_], indices)
        indices = sorted(indices)
        indices.insert(0, 0)
        indices.append(max_)

        output = []
        for i in indices:
            if i not in output:
                output.append(i)
        return output
    # ------------------------------------------------------------------------

    # get axes
    axes = mode.split('-')
    a0 = axes[0]

    # get indices along first axis
    indices = get_indices(image[:, :, channel], a0)
    segments = cut(image, indices, axis=a0)

    # if only one axis return segments
    output = {}
    if len(axes) == 1:
        if mode == 'vertical':
            output = {(x, 0): item for x, item in enumerate(segments)}
        else:
            yl = len(segments) - 1
            output = {(0, yl - y): item for y, item in enumerate(segments)}
        return output

    # otherwise chop each segement according to its edge detect indices
    a1 = axes[1]
    xl = len(segments) - 1
    for x, segment in enumerate(segments):
        indices = get_indices(segment[:, :, channel], a1)
        images = cut(segment, indices, axis=a1)
        yl = len(images) - 1
        for y, image in enumerate(images):
            if a0 == 'vertical':
                output[(x, yl - y)] = image
            else:
                output[(y, xl - x)] = image
    return output


def warp(
    image,
    angle=0,
    translate_x=0,
    translate_y=0,
    scale=1,
    inverse=False,
    matrix=None
):
    # type: (Image, float, float, float, float, bool, OptArray) -> Image
    '''
    Warp image by given parameters or warp matrix.
    If matrix is given, parameters are ignored.

    Args:
        image (Image): Image.
        angle (float, optional): Rotation in degrees. Default: 0.
        translate_x (float, optional): X translation. Default: 0.
        translate_y (float, optional): Y translation. Default: 0.
        scale (float, optional): Scale factor. Default: 1.
        inverse (bool, optional): Inverse transformation. Default: False.
        matrix (numpy.ndarray, optional): Warp matrix. Default: None.

    Raises:
        EnforceError: If image is not an instance of Image.
        EnforceError: If matrix is not None or an instance of np.ndarray.
        EnforceError: If matrix is not a 3 x 3 matrix.

    Returns:
        Image: Warped image.
    '''
    Enforce(image, 'instance of', Image)

    if matrix is None:
        matrix = get_warp_matrix(angle, translate_x, translate_y, scale)
    Enforce(matrix, 'instance of', np.ndarray)
    Enforce(matrix.shape, '==', (3, 3))
    # --------------------------------------------------------------------------

    # convert to float
    bit_depth = image.bit_depth
    channels = image.channels
    array = image.to_bit_depth(BitDepth.FLOAT32).data
    flag = cv2.WARP_INVERSE_MAP if inverse else 0
    array = cv2.warpPerspective(
        array, matrix, dsize=image.width_and_height, flags=flag
    )

    # convert to original bit depth
    output = Image.from_array(array).to_bit_depth(bit_depth)

    # set original channel names
    output = output.set_channels(channels)
    return output


def get_warp_matrix(angle=0, translate_x=0, translate_y=0, scale=1):
    # type: (float, float, float, float) -> np.ndarray
    '''
    Create 3 x 3 warp matrix.

    Args:
        angle (float, optional): Rotation in degrees. Default: 0.
        translate_x (float, optional): X translation. Default: 0.
        translate_y (float, optional): Y translation. Default: 0.
        scale (float, optional): Scale factor. Default: 1.

    Returns:
        ndarray: 3 x 3 warp matrix.
    '''
    t = np.radians(angle)
    x = translate_x
    y = translate_y
    s = scale
    return np.array([
        [math.cos(t) * s, -math.sin(t) * s, x],
        [math.sin(t) * s, math.cos(t) * s, y],
        [0, 0, 1]
    ])



from typing import Optional, Union
from pathlib import Path

import numpy as np

from cv_depot.core.enum import Anchor
from cv_depot.core.color import BasicColor, Color
# ------------------------------------------------------------------------------

AnyAnchor = Union[Anchor, str]
AnyColor = Union[Color, BasicColor, str]
Filepath = Union[str, Path]
OptArray = Optional[np.ndarray]



from typing import Any, Optional, Union  # noqa F401
from numpy.typing import NDArray  # noqa F401

import json
import math
from copy import copy
from enum import Enum

from lunchbox.enforce import Enforce
import numpy as np

from cv_depot.core.enum import BitDepth
# ------------------------------------------------------------------------------


'''
Color is a fundamentally a vector of length c, where c is number of channels.
Colors can only be of bit depths supported by the BitDepth class.
'''


class BasicColor(Enum):
    '''
    A convenience enum for basic, common color vectors.
    Legal colors include:

        * BLACK
        * WHITE
        * GREY
        * RED
        * GREEN
        * BLUE
        * YELLOW
        * MAGENTA
        * CYAN
    '''
    BLACK = ('#000000', [0.0], [0.0, 0.0, 0.0])
    WHITE = ('#FFFFFF', [1.0], [1.0, 1.0, 1.0])
    GREY = ('#808080', [0.5], [0.5, 0.5, 0.5])
    RED = ('#FF0000', None, [1.0, 0.0, 0.0])
    GREEN = ('#00FF00', None, [0.0, 1.0, 0.0])
    BLUE = ('#0000FF', None, [0.0, 0.0, 1.0])
    YELLOW = ('#FFFF00', None, [1.0, 1.0, 0.0])
    MAGENTA = ('#FF00FF', None, [1.0, 0.0, 1.0])
    CYAN = ('#00FFFF', None, [0.0, 1.0, 1.0])

    # henanigans
    BG = ('#242424', [0.141], [0.141, 0.141, 0.141])
    BLUE1 = ('#5F95DE', None, [0.373, 0.584, 0.871])
    BLUE2 = ('#93B6E6', None, [0.576, 0.714, 0.902])
    CYAN1 = ('#7EC4CF', None, [0.494, 0.769, 0.812])
    CYAN2 = ('#B6ECF3', None, [0.714, 0.925, 0.953])
    DARK1 = ('#040404', [0.016], [0.016, 0.016, 0.016])
    DARK2 = ('#141414', [0.078], [0.078, 0.078, 0.078])
    DIALOG1 = ('#444459', None, [0.267, 0.267, 0.349])
    DIALOG2 = ('#5D5D7A', None, [0.365, 0.365, 0.478])
    GREEN1 = ('#8BD155', None, [0.545, 0.82, 0.333])
    GREEN2 = ('#A0D17B', None, [0.627, 0.82, 0.482])
    GREY1 = ('#343434', [0.204], [0.204, 0.204, 0.204])
    GREY2 = ('#444444', [0.267], [0.267, 0.267, 0.267])
    LIGHT1 = ('#A4A4A4', [0.643], [0.643, 0.643, 0.643])
    LIGHT2 = ('#F4F4F4', [0.957], [0.957, 0.957, 0.957])
    ORANGE1 = ('#EB9E58', None, [0.922, 0.62, 0.345])
    ORANGE2 = ('#EBB483', None, [0.922, 0.706, 0.514])
    PURPLE1 = ('#C98FDE', None, [0.788, 0.561, 0.871])
    PURPLE2 = ('#AC92DE', None, [0.675, 0.573, 0.871])
    RED1 = ('#F77E70', None, [0.969, 0.494, 0.439])
    RED2 = ('#DE958E', None, [0.871, 0.584, 0.557])
    YELLOW1 = ('#E8EA7E', None, [0.91, 0.918, 0.494])
    YELLOW2 = ('#E9EABE', None, [0.914, 0.918, 0.745])

    def __init__(self, hexidecimal, one_channel, three_channel):
        # type: (str, list[float], list[float]) -> None
        '''
        Args:
            hexidecimal (str): Hexidecimal representation of color.
            one_channel (list[float]): List with single float.
            three_channel (list[float]): List with three floats.

        Returns:
            BasicColor: BasicColor instance.
        '''
        self._hexidecimal = hexidecimal
        self._one_channel = one_channel
        self._three_channel = three_channel

    @staticmethod
    def _get_color(
        value,  # type: Union[str, BasicColor, list[int], list[float]]
        attr    # type: str
    ):          # type: (...) -> BasicColor
        '''
        Finds BasicColor instance given a value and an attribute name.

        Args:
            value (str or BasicColor): BasicColor value to be looked up.
            attr (str): Attribute name to be used in lookup.

        Raises:
            ValueError: If no BasicColor corresponds to given value.

        Returns:
            BasicColor: BasicColor instance.
        '''
        lut = {
            json.dumps(getattr(x, attr)): x for x in BasicColor
        }  # type: dict[str, BasicColor]

        if attr == 'string':
            value = str(value).lower()
        elif attr == 'hexidecimal':
            value = str(value).upper()

        key = json.dumps(value)
        if key not in lut.keys():
            msg = f'{value} is not a legal color.'
            raise ValueError(msg)
        return lut[key]

    @staticmethod
    def from_string(value):
        # type: (str) -> BasicColor
        '''
        Contructs a BasicColor instance from a given string.

        Args:
            value (str): Name of color.

        Returns:
            BasicColor: BasicColor instance.
        '''
        return BasicColor._get_color(value, 'string')

    @staticmethod
    def from_list(value):
        # type: (list) -> BasicColor
        '''
        Contructs a BasicColor instance from a given list of floats or ints.

        Args:
            value (list): BasicColor as list of numbers.

        Returns:
            BasicColor: BasicColor instance.
        '''
        value = list(map(float, value))
        if len(value) == 1:
            return BasicColor._get_color(value, 'one_channel')
        elif len(value) == 3:
            return BasicColor._get_color(value, 'three_channel')

        msg = f'Invalid color value {value}. Must be 1 or 3 channels.'
        raise ValueError(msg)

    @staticmethod
    def from_list_8_bit(value):
        # type: (list[int]) -> BasicColor
        '''
        Contructs a BasicColor instance from a given list of ints between 0 and
        255.

        Args:
            value (list): BasicColor as list of numbers.

        Returns:
            BasicColor: BasicColor instance.
        '''
        value_ = [x / 255 for x in value]
        return BasicColor.from_list(value_)

    @staticmethod
    def from_hexidecimal(value):
        # type: (str) -> BasicColor
        '''
        Contructs a BasicColor instance from a given hexidecimal string.

        Args:
            value (str): Hexidecimal value of color.

        Returns:
            BasicColor: BasicColor instance.
        '''
        return BasicColor._get_color(value, 'hexidecimal')

    def __repr__(self):
        # type: () -> str
        return f'''
<BasicColor.{self.name.upper()}>
             string: {self.string}
        hexidecimal: {self.hexidecimal}
        one_channel: {self.one_channel}
      three_channel: {self.three_channel}
  one_channel_8_bit: {self.one_channel_8_bit}
three_channel_8_bit: {self.three_channel_8_bit}'''[1:]

    @property
    def hexidecimal(self):
        # type: () -> str
        '''
        str: Hexidecimal representation of color.
        '''
        return copy(self._hexidecimal)

    @property
    def one_channel(self):
        # type: () -> list[float]
        '''
        list[float]: One channel, floating point representation of color.
        '''
        return copy(self._one_channel)

    @property
    def three_channel(self):
        # type: () -> list[float]
        '''
        list[float]: Three channel, floating point representation of color.
        '''
        return copy(self._three_channel)

    @property
    def one_channel_8_bit(self):
        # type: () -> list[int]
        '''
        list[int]: One channel, 8 bit representation of color.
        '''
        if self.one_channel is None:
            return None
        return [int(math.ceil(self.one_channel[0] * 255))]

    @property
    def three_channel_8_bit(self):
        # type: () -> list[int]
        '''
        list[int]: Three channel, 8 bit representation of color.
        '''
        return [int(math.ceil(x * 255)) for x in self.three_channel]

    @property
    def string(self):
        # type: () -> str
        '''
        string: String representation of color.
        '''
        return self.name.lower()
# ------------------------------------------------------------------------------


class Color:
    '''
    Makes working with color vectors easy.
    '''
    bit_depth = BitDepth.FLOAT32

    def __init__(self, data):
        # type: (NDArray) -> None
        '''
        Constructs a Color instance.

        Args:
            data (numpy.NDArray): A 1 dimensional numpy array of shape (n,).

        Raises:
            TypeError: If data is not a numpy array.
            AttributeError: If data's number of dimensions is not 1.

        Returns:
            Color: Color instance.
        '''
        if not isinstance(data, np.ndarray):
            msg = 'Data must be a numpy array.'
            raise TypeError(msg)

        if data.ndim != 1:
            msg = 'Arrays must be one dimensional, so that its shape is (n,). '
            msg += f'Given array has {data.ndim} dimensions and a shape of '
            msg += f'{data.shape}.'
            raise AttributeError(msg)

        # checks for legal bit depth
        bit_depth = BitDepth.from_dtype(data.dtype)
        data = data.astype(self.bit_depth.dtype)

        # 8 bit conversion
        if bit_depth is BitDepth.INT8:
            data = (data + 127) / 255
        elif bit_depth is BitDepth.UINT8:
            data = data / 255

        self._data = data

    @staticmethod
    def from_array(data, num_channels=None, fill_value=0):
        # type: (NDArray, Optional[int], float) -> Color
        '''
        Contructs a Color instance from a given numpy array.

        Args:
            data (numpy.NDArray): Numpy array.
            num_channels (int, optional): Number of desired channels in the
                Color instance. Number of channels equals len(data) if set to
                None. Default: None.
            fill_value (float, optional): Value used to fill additional
                channels. Default: 0.

        Returns:
            Color: Color instance of given data.
        '''
        if num_channels is not None:
            Enforce(num_channels, 'instance of', int)
            Enforce(num_channels, '>=', 1)

            diff = num_channels - len(data)
            if diff > 0:
                buff = np.ones(diff, data.dtype) * fill_value
                data = np.append(data, buff)
            data = data[:num_channels]
        return Color(data)

    @staticmethod
    def from_list(
        data, num_channels=None, fill_value=0, bit_depth=BitDepth.FLOAT32
    ):
        # type: (NDArray, Optional[int], float, BitDepth) -> Color
        '''
        Contructs a Color instance from a given list of ints or floats.

        Args:
            data (list): List of ints or floats.
            num_channels (int, optional): Number of desired channels in the
                Color instance. Number of channels equals len(data) if set to
                None. Default: None.
            fill_value (float, optional): Value used to fill additional
                channels. Default: 0.
            bit_depth (BitDepth, optional): Bit depth of given list.
                Default: BitDepth.FLOAT32.

        Returns:
            Color: Color instance of given data.
        '''
        data = np.array(data, dtype=bit_depth.dtype)
        return Color.from_array(
            data, num_channels=num_channels, fill_value=fill_value
        )

    @staticmethod
    def from_basic_color(data, num_channels=3, fill_value=0.0):
        # type: (BasicColor, int, float) -> Color
        '''
        Contructs a Color instance from a BasicColor enum.

        Args:
            data (BasicColor): BasicColor enum.
            num_channels (int, optional): Number of desired channels in the
                Color instance. Default: 3.
            fill_value (float, optional): Value used to fill additional
                channels. Default: 0.

        Raises:
            EnforceError: If num_channels is less than 1.
            EnforceError: If no shape is one channel and no one channel
                equivalent of the given color could be found.

        Returns:
            Color: Color instance of given data.
        '''
        msg = 'num_channels must be greater than or equal to 1. {a} < {b}.'
        Enforce(num_channels, '>=', 1, message=msg)

        output = data  # type: Any
        if num_channels == 1:
            msg = f'No one channel equivalent found for given color: {data}.'
            Enforce(data.one_channel, '!=', None, message=msg)
            output = data.one_channel
        else:
            output = data.three_channel

        return Color.from_list(
            output,
            num_channels=num_channels,
            fill_value=fill_value,
            bit_depth=BitDepth.FLOAT32
        )

    def __repr__(self):
        # type: () -> str
        '''
        Represention of Color instance includes:

            * values - actual numpy array
            * bit_depth - always FLOAT32
            * num_channels - number of color channels
            * name - displays only if the color values corresponds to a
              BasicColor
        '''
        output = [
            '<Color>',
            f'      values: {self._data}',
            f'   bit_depth: {self.bit_depth.name.upper()}',
            f'num_channels: {self.num_channels}',
        ]  # type: Any

        try:
            items = self.to_array().tolist()
            name = BasicColor.from_list(items).string  # type: ignore
            output.append(f'        name: {name}')
        except Exception:
            pass

        output = '\n'.join(output)
        return output

    @property
    def num_channels(self):
        # type: () -> int
        '''
        Number of channels in color vector.
        '''
        return len(self._data)

    def to_array(self, bit_depth=BitDepth.FLOAT32):
        # type: (BitDepth) -> NDArray
        '''
        Returns color as a numpy array at a given bit depth.

        Args:
            bit_depth (BitDepth, optional): Bit depth of output.
                Default: BitDepth.FLOAT32.

        Returns:
            numpy.NDArray: Color vector as numpy array.
        '''
        output = self._data
        if bit_depth is BitDepth.INT8:
            output = np.ceil(output * 255) - 128
        elif bit_depth is BitDepth.UINT8:
            output = np.ceil(output * 255)
        return output.astype(bit_depth.dtype)



from typing import Any, Optional, Union  # noqa F401
from cv_depot.core.types import AnyColor  # noqa F401

import logging

from lunchbox.enforce import Enforce
from lunchbox.stopwatch import StopWatch
import cv2
import numpy as np

from cv_depot.core.image import BitDepth, Image
from cv_depot.core.color import BasicColor, Color

LOGGER = logging.getLogger(__name__)
# ------------------------------------------------------------------------------


def gamma(image, value=1.0):
    # type: (Image, float) -> Image
    '''
    Apply gamma correction to given image.

    Args:
        image (Image): Image to be modified.
        value (int): Gamma value. Default: 1.0.

    Raises:
        EnforceError: If image is not an instance of Image.
        EnforceError: If value is less than 0.

    Returns:
        Image: Gamma adjusted image.
    '''
    Enforce(image, 'instance of', Image)
    Enforce(value, '>=', 0)
    # --------------------------------------------------------------------------

    bit_depth = image.bit_depth
    array = image.to_bit_depth(BitDepth.FLOAT32).data
    array = array**(1 / value)
    output = Image.from_array(array).to_bit_depth(bit_depth)
    return output


def canny_edges(image, size=0):
    # type: (Image, int) -> Image
    '''
    Apply a canny edge detection to given image.

    Args:
        image (Image): Image.
        size (int, optional): Amount of dilation applied to result. Default: 0.

    Raises:
        EnforceError: If image is not an instance of Image.
        EnforceError: If size is not an integer >= 0.

    Returns:
        Image: Edge detected image.
    '''
    Enforce(image, 'instance of', Image)
    Enforce(size, 'instance of', int)
    Enforce(size, '>=', 0)
    # --------------------------------------------------------------------------

    img = image.to_bit_depth(BitDepth.UINT8).data
    img = cv2.Canny(img, 0, 0)
    kernel = np.ones((3, 3), np.uint8)
    img = cv2.dilate(img, kernel, iterations=size)
    output = Image.from_array(img).to_bit_depth(image.bit_depth)
    return output


def tophat(image, amount, kind='open'):
    # type: (Image, int, str) -> Image
    '''
    Apply tophat morphology operation to given image.

    Args:
        image (Image): Image to be modified.
        amount (int): Amount of tophat operation.
        kind (str, optional): Kind of operation to be performed.
            Options include: open, close. Default: open.

    Raises:
        EnforceError: If image is not an instance of Image.
        EnforceError: If amount is less than 0.
        EnforceError: If kind is not one of: open, close.

    Returns:
        Image: Image with tophat operation applied.
    '''
    Enforce(image, 'instance of', Image)
    Enforce(amount, '>=', 0)
    msg = 'Illegal tophat kind: {a}. Legal tophat kinds: {b}.'
    Enforce(kind, 'in', ['open', 'close'], message=msg)
    # --------------------------------------------------------------------------

    lut = dict(open=cv2.MORPH_CLOSE, close=cv2.MORPH_OPEN)
    opt = lut[kind]
    bit_depth = image.bit_depth
    img = image.to_bit_depth(BitDepth.FLOAT32).data
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3))
    arr = cv2 \
        .morphologyEx(img, opt, kernel, iterations=amount) \
        .astype(np.float32)
    output = Image.from_array(arr).to_bit_depth(bit_depth)
    return output


def linear_lookup(lower=0, upper=1):
    # type: (float, float) -> np.vectorize
    r'''
    Generates a linear lookup table with an upper and lower shoulder.

    .. image:: images/linear_lut.png

    Args:
        lower (float, optional): Lower shoulder value. Default: 0.
        upper (float, optional): Upper shoulder value. Default: 1.

    Returns:
        numpy.vectorize: Anonymous function that applies lut elementwise to a
            given numpy array.
    '''
    lut = lambda x: min(max((x - lower), 0) * (1 / (upper - lower)), 1)
    return np.vectorize(lut)


def linear_smooth(image, blur=3, lower=0, upper=1):
    # type: (Image, int, float, float) -> Image
    '''
    Blur given image then apply linear thresholding it.

    Args:
        image (Image): Image matte to be smoothed.
        blur (int, optional): Size of blur. Default: 3.
        lower (float, optional): Lower shoulder value. Default: 0.
        upper (float, optional): Upper shoulder value. Default: 1.

    Raises:
        EnforceError: If image is not an instance of Image.
        EnforceError: If blur is less than 0.
        EnforceError: If lower or upper is less than 0.
        EnforceError: If lower or upper is greater than 1.
        EnforceError: If lower is greater than upper.

    Returns:
        Image: Smoothed image.
    '''
    Enforce(image, 'instance of', Image)
    Enforce(blur, '>=', 0)
    Enforce(lower, '>=', 0)
    Enforce(lower, '<=', 1)
    Enforce(upper, '>=', 0)
    Enforce(upper, '<=', 1)
    Enforce(lower, '>=', 0)
    msg = 'Lower bound cannot be greater than upper bound. {a} > {b}'
    Enforce(lower, '<=', upper, message=msg)
    # --------------------------------------------------------------------------

    bit_depth = image.bit_depth
    img = image.to_bit_depth(BitDepth.FLOAT32).data
    img = cv2.blur(img, (blur, blur))
    lut = linear_lookup(lower=lower, upper=upper)
    img = lut(img).astype(np.float32)
    output = Image.from_array(img).to_bit_depth(bit_depth)
    return output


def key_exact_color(image, color, channel='a', invert=False):
    # type: (Image, AnyColor, str, bool) -> Image
    '''
    Keys given image according to the color of its pixels values.
    Where that pixel color exactly matches the given color, the mask channel
    will be 1, otherwise it will be 0.

    Args:
        image (Image): Image to be evaluated.
        color (Color or BasicColor): Color to be used for masking.
        channel (str, optional): Mask channel name. Default: a.
        invert (bool, optional): Whether to invert the mask. Default: False.

    Raises:
        EnforceError: If image is not an Image instance.
        EnforceError: If channel is not a string.
        EnforceError: If invert is not a boolean.
        EnforceError: If RGB is not found in image channels.

    Returns:
        Image: Image with mask channel.
    '''
    Enforce(image, 'instance of', Image)
    Enforce(channel, 'instance of', str)
    Enforce(invert, 'instance of', bool)
    # --------------------------------------------------------------------------

    # get color
    if isinstance(color, str):
        color = BasicColor.from_string(color)
    if isinstance(color, BasicColor):
        color = Color.from_basic_color(color)
    clr = color.to_array()

    # determine num channels
    rgb = list('rgb')
    img = image.to_bit_depth(BitDepth.FLOAT32)
    if image.num_channels == 1:
        x = img.data[..., np.newaxis]
        x = np.concatenate([x, x, x], axis=2)
        img = Image.from_array(x)
    else:
        diff = sorted(list(set(rgb).difference(image.channels)))
        msg = f'{diff} not found in image channels. '
        msg += f'Given channels: {image.channels}.'
        Enforce(len(diff), '==', 0, message=msg)

    # create mask
    mask = np.equal(clr, img[:, :, rgb].data)
    mask = np.apply_along_axis(all, 2, mask) \
        .astype(np.float32)[..., np.newaxis]
    if invert:
        mask = -1 * mask + 1

    # add mask to image
    chans = image.channels
    if image.num_channels == 1:
        arr = img[:, :, 'r'].data[..., np.newaxis]
    else:
        chans = list(filter(lambda x: x != channel, img.channels))
        arr = img[:, :, chans].data

    arr = np.concatenate([arr, mask], axis=2)
    output = Image.from_array(arr).to_bit_depth(image.bit_depth)
    output = output.set_channels(chans + [channel])
    return output


def kmeans(
    image,                 # type: Image
    num_centroids=10,      # type: int
    centroids=None,        # type: Optional[list[tuple[int, int, int]]]
    max_iterations=100,    # type: int
    accuracy=1.0,          # type: float
    epochs=10,             # type: int
    seeding='random',      # type: str
    generate_report=False  # type: bool
):                         # type: (...) -> Union[Image, tuple[Image, dict]]
    '''
    Applies k-means to the given image.

    Args:
        image (Image): Image instance.
        num_centroids (int, optional): Number of centroids to use. Default: 10.
        centroids (list, optional): List of triplets. Default: None.
        max_iterations (int, optional): Maximum number of k-means updates
            allowed per centroid. Default: 100.
        accuracy (float, optional): Minimum accuracy required of clusters.
        epochs (int, optional): Number of times algorithm is applied with
            different initial labels. Default: 10.
        seeding (str, optional): How intial centroids are generated. Default:
            random. Options include: random, pp_centers.
        generate_report (bool, optional): If true returns report in addition to
            image. Default: False.

    Raises:
        EnforceError: If image is not an Image instance.
        ValueError: If invalid seeding option is given.

    Returns:
        Image or tuple[Image, dict]: K-means image or K-means image and K-means
            report.
    '''
    Enforce(image, 'instance of', Image)

    stopwatch = StopWatch()
    stopwatch.start()
    # --------------------------------------------------------------------------

    source_bit_depth = image.bit_depth
    data = image.to_bit_depth(BitDepth.FLOAT32)\
        .data.reshape((-1, image.num_channels))

    seed = None
    if seeding == 'random':
        seed = cv2.KMEANS_RANDOM_CENTERS
    elif seeding == 'pp_centers':
        seed = cv2.KMEANS_PP_CENTERS
    else:
        msg = f'{seeding} is not a valid seeding option. Options include: '
        msg += '[random, pp_centers].'
        raise ValueError(msg)

    # terminate centroid updates when max iteration or min accuracy is achieved
    crit = cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER

    if centroids is not None:
        num_centroids = len(centroids)

    compactness, labels, centroids = cv2.kmeans(
        data=data,
        K=num_centroids,
        bestLabels=centroids,
        criteria=(crit, max_iterations, accuracy),
        attempts=epochs,
        flags=seed,
    )  # type: ignore

    centroids_ = centroids  # type: Any
    output = np.float32(centroids_)[labels.flatten()]  # type: ignore
    output = output.reshape(image.data.shape)
    output = Image.from_array(output).to_bit_depth(source_bit_depth)

    if generate_report:
        report = dict(
            compactness=compactness,
            labels=labels,
            centroids=centroids
        )
        stopwatch.stop()
        LOGGER.warning(f'KmeansRuntime: {stopwatch.human_readable_delta}.')
        return output, report

    stopwatch.stop()
    LOGGER.warning(f'Kmeans Runtime: {stopwatch.human_readable_delta}.')
    return output



from typing import Any  # noqa F401

from enum import Enum
import re

from lunchbox.enforce import Enforce
import numpy as np
# ------------------------------------------------------------------------------


'''
The enum module contains Enum classes for manging aspects of imagery such as bit
depths and video codecs.
'''


class EnumBase(Enum):
    def __repr__(self):
        # type: () -> str
        '''
        str: String representation of enum.
        '''
        return f'{self.__class__.__name__}.{self.name.upper()}'

    @classmethod
    def from_string(cls, string):
        # type: (str) -> EnumBase
        '''
        Constructs an enum instance from a given string.

        Args:
            string (int): Enum string.

        Raises:
            EnforceError: If value given is not a string.
            EnforceError: If no EnumBase type can be found for given string.

        Returns:
            EnumBase: Enum instance.
        '''
        msg = 'Value given is not a string. {a} != {b}.'
        Enforce(string, 'instance of', str, message=msg)

        lut = {x.name: x for x in cls.__members__.values()}
        string = string.upper().replace('-', '_')

        msg = '{a} is not a ' + cls.__name__ + ' option. '
        msg += f'Options: {sorted(lut.keys())}.'
        Enforce(string, 'in', lut.keys(), message=msg)

        return lut[string]


# BITDEPTH----------------------------------------------------------------------
class BitDepth(EnumBase):
    '''
    Legal bit depths.

    Includes:

        * FLOAT16
        * FLOAT32
        * UINT8
        * INT8
    '''
    FLOAT16 = (np.float16, 16, True, float)
    FLOAT32 = (np.float32, 32, True, float)
    UINT8 = (np.uint8, 8, False, int)
    INT8 = (np.int8, 8, True, int)

    def __init__(self, dtype, bits, signed, type_):
        # type: (Any, int, bool, type) -> None
        '''
        Args:
            dtype (numpy.type): Numpy datatype.
            bits (int): Number of bits per channel.
            signed (bool): Whether channel scalars are signed.
            type_ (type): Python type of scalar. Options include: [int, float].

        Returns:
            BitDepth: BitDepth instance.
        '''
        self.dtype = dtype  # type: ignore
        self.bits = bits
        self.signed = signed
        self.type_ = type_

    def __repr__(self):
        # type: () -> str
        return f'BitDepth.{self.name.upper()}'

    @staticmethod
    def from_dtype(dtype):
        # type: (Any) -> BitDepth
        '''
        Construct a BitDepth instance from a given numpy datatype.

        Args:
            dtype (numpy.type): Numpy datatype. Options include:
                [float16, float32, uint8, int8].

        Raises:
            TypeError: If invlaid dtype is given.

        Returns:
            BitDepth: BitDepth instance of given type.
        '''
        if dtype == np.float16:
            return BitDepth.FLOAT16
        elif dtype == np.float32:
            return BitDepth.FLOAT32
        elif dtype == np.uint8:
            return BitDepth.UINT8
        elif dtype == np.int8:
            return BitDepth.INT8

        # needed because of numpy malarkey with __name__
        if hasattr(dtype, '__name__'):
            dtype = dtype.__name__
        msg = f'{dtype} is not a supported bit depth.'
        raise TypeError(msg)


# IMAGE-------------------------------------------------------------------------
class ImageFormat(Enum):
    '''
    Legal image formats.

    Includes:

        * EXR
        * PNG
        * JPEG
        * TIFF
    '''
    EXR = (
        'exr', [BitDepth.FLOAT16, BitDepth.FLOAT32], list('rgba') + ['...'],
        1023, True
    )
    PNG = ('png', [BitDepth.UINT8], list('rgba'), 4, False)
    JPEG = ('jpeg', [BitDepth.UINT8], list('rgb'), 3, False)
    TIFF = (
        'tiff', [BitDepth.INT8, BitDepth.UINT8, BitDepth.FLOAT32],
        list('rgba') + ['...'], 500, False
    )

    def __init__(self, extension, bit_depths, channels, max_channels,
                 custom_metadata):
        # type: (str, list[BitDepth], list[str], int, bool) -> None
        '''
        Args:
            extension (str): Name of file extension.
            bit_depths (list[BitDepth]): Supported bit depths.
            channels (list[str]): Supported channels.
            max_channels (int): Maximum number of channels supported.
            custom_metadata (bool): Custom metadata support.

        Returns:
            ImageFormat: ImageFormat instance.
        '''
        self.extension = extension
        self.bit_depths = bit_depths
        self.channels = channels
        self.max_channels = max_channels
        self.custom_metadata = custom_metadata

    def __repr__(self):
        # type: () -> str
        return f'''
<ImageFormat.{self.name.upper()}>
      extension: {self.extension}
     bit_depths: {[x.name for x in self.bit_depths]}
       channels: {self.channels}
   max_channels: {self.max_channels}
custom_metadata: {self.custom_metadata}'''[1:]

    @staticmethod
    def from_extension(extension):
        '''
        Construct an ImageFormat instance for a given file extension.

        Args:
            extension (str): File extension.

        Raises:
            TypeError: If extension is illegal.

        Returns:
            ImageFormat: ImageFormat instance of given extension.
        '''
        exr_re = r'^\.?exr$'
        png_re = r'^\.?png$'
        jpeg_re = r'^\.?jpe?g$'
        tiff_re = r'^\.?tiff?$'

        if re.search(exr_re, extension, re.I):
            return ImageFormat.EXR

        elif re.search(png_re, extension, re.I):
            return ImageFormat.PNG

        elif re.search(jpeg_re, extension, re.I):
            return ImageFormat.JPEG

        elif re.search(tiff_re, extension, re.I):
            return ImageFormat.TIFF

        msg = f'ImageFormat not found for given extension: {extension}'
        raise TypeError(msg)


# VIDEO-------------------------------------------------------------------------
class VideoFormat(Enum):
    '''
    Legal video formats.

    Includes:

        * MP4
        * MPEG
        * MOV
        * M4V
    '''
    MP4 = ('mp4', [BitDepth.UINT8], list('rgb'), 3, False)
    MPEG = ('mpeg', [BitDepth.UINT8], list('rgb'), 3, False)
    MOV = ('mov', [BitDepth.UINT8], list('rgb'), 3, False)
    M4V = ('m4v', [BitDepth.UINT8], list('rgb'), 3, False)

    def __init__(self, extension, bit_depths, channels, max_channels,
                 custom_metadata):
        # type: (str, list[BitDepth], list[str], int, bool) -> None
        '''
        Args:
            extension (str): Name of file extension.
            bit_depths (list[BitDepth]): Supported bit depths.
            channels (list[str]): Supported channels.
            max_channels (int): Maximum number of channels supported.
            custom_metadata (bool): Custom metadata support.

        Returns:
            VideoFormat: VideoFormat instance.
        '''
        self.extension = extension
        self.bit_depths = bit_depths
        self.channels = channels
        self.max_channels = max_channels
        self.custom_metadata = custom_metadata

    def __repr__(self):
        # type: () -> str
        return f'''
<VideoFormat.{self.name.upper()}>
      extension: {self.extension}
     bit_depths: {[x.name for x in self.bit_depths]}
       channels: {self.channels}
   max_channels: {self.max_channels}
custom_metadata: {self.custom_metadata}'''[1:]

    @staticmethod
    def from_extension(extension):
        '''
        Construct an VideoFormat instance for a given file extension.

        Args:
            extension (str): File extension.

        Raises:
            TypeError: If extension is invalid.

        Returns:
            VideoFormat: VideoFormat instance of given extension.
        '''
        mp4_re = r'^\.?mp4$'
        mpeg_re = r'^\.?mpe?g$'
        mov_re = r'^\.?mov$'
        m4v_re = r'^\.?m4v$'

        if re.search(mp4_re, extension, re.I):
            return VideoFormat.MP4

        if re.search(mpeg_re, extension, re.I):
            return VideoFormat.MPEG

        if re.search(mov_re, extension, re.I):
            return VideoFormat.MOV

        if re.search(m4v_re, extension, re.I):
            return VideoFormat.M4V

        msg = f'VideoFormat not found for given extension: {extension}'
        raise TypeError(msg)
# ------------------------------------------------------------------------------


class VideoCodec(Enum):
    '''
    Legal video codecs.

    Includes:

        * H264
        * H265
    '''
    H264 = ('h264', 'h264')
    H265 = ('h265', 'hevc')

    def __init__(self, string, ffmpeg_code):
        # type: (str, str) -> None
        '''
        Args:
            string (str): String representation of codec.
            ffmpeg_code (str): FFMPEG code.
        '''
        self.string = string
        self.ffmpeg_code = ffmpeg_code

    def __repr__(self):
        # type: () -> str
        return f'''
<VideoCodec.{self.name.upper()}>
  string: {self.string}
  ffmpeg_code: {self.ffmpeg_code}'''[1:]


# DIRECTION---------------------------------------------------------------------
class Direction(EnumBase):
    '''
    Legal directions.

    Includes:

        * TOP
        * BOTTOM
        * LEFT
        * RIGHT
    '''
    TOP = ('top')
    BOTTOM = ('bottom')
    LEFT = ('left')
    RIGHT = ('right')


class Anchor(EnumBase):
    '''
    Legal anchors.

    Includes:

        * TOP_LEFT
        * TOP_CENTER
        * TOP_RIGHT
        * CENTER_LEFT
        * CENTER_CENTER
        * CENTER_RIGHT
        * BOTTOM_LEFT
        * BOTTOM_CENTER
        * BOTTOM_RIGHT
    '''
    TOP_LEFT = ('top', 'left')
    TOP_CENTER = ('top', 'center')
    TOP_RIGHT = ('top', 'right')
    CENTER_LEFT = ('center', 'left')
    CENTER_CENTER = ('center', 'center')
    CENTER_RIGHT = ('center', 'right')
    BOTTOM_LEFT = ('bottom', 'left')
    BOTTOM_CENTER = ('bottom', 'center')
    BOTTOM_RIGHT = ('bottom', 'right')



from typing import Any, Optional, Tuple, Union  # noqa F401
from numpy.typing import NDArray  # noqa F401
from cv_depot.core.types import Filepath  # noqa F401

from copy import deepcopy
from itertools import combinations, chain
from pathlib import Path
import os
import re

from lunchbox.enforce import Enforce
from openexr_tools.enum import ImageCodec
import numpy as np
import openexr_tools.tools as exrtools
import PIL.Image as pil

from cv_depot.core.enum import BitDepth, ImageFormat
from cv_depot.core.viewer import ImageViewer
import cv_depot.core.tools as cvt
# ------------------------------------------------------------------------------


def _has_super_darks(image):
    # type: (Image) -> bool
    '''
    Determines if given image has values below 0.0

    Args:
        image (Image): Image instance.

    Raises:
        EnforceError: If image is not an Image instance.

    Returns:
        bool: Presence of super darks.
    '''
    Enforce(image, 'instance of', Image)
    return bool(image.data.min() < 0.0)


def _has_super_brights(image):
    # type: (Image) -> bool
    '''
    Determines if given image has values above 1.0

    Args:
        image (Image): Image instance.

    Raises:
        EnforceError: If image is not an Image instance.

    Returns:
        bool: Presence of super brights.
    '''
    Enforce(image, 'instance of', Image)
    return bool(image.data.max() > 1.0)


class Image():
    '''
    Class for reading, writing, converting and displaying properties of images.
    '''
    @staticmethod
    def from_array(array):
        # type: (NDArray) -> Image
        '''
        Construct an Image instance from a given numpy array.

        Args:
            array (numpy.NDArray): Numpy array.

        Returns:
            Image: Image instance of given numpy array.
        '''
        # enforce bit depth compliance
        BitDepth.from_dtype(array.dtype)
        return Image(array.copy(), {}, None, allow=True)

    @staticmethod
    def from_pil(image):
        # type: (pil.Image) -> Image
        '''
        Construct an Image instance from a given PIL Image.

        Args:
            image (pil.Image): PIL Image.

        Returns:
            Image: Image instance of a given PIL Image.
        '''
        return Image.from_array(np.array(image))

    @staticmethod
    def read(filepath):
        # type: (Filepath) -> Image
        '''
        Constructs an Image instance given a full path to an image file.

        Args:
            filepath (str or Path): Image filepath.

        Raises:
            FileNotFoundError: If file could not be found on disk.
            TypeError: If filepath is not a str or Path.

        Returns:
            Image: Image instance of given file.
        '''
        metadata = {}  # type: dict[str, Any]
        format_ = None

        if isinstance(filepath, Path):
            filepath = filepath.absolute().as_posix()

        if isinstance(filepath, str):
            if not os.path.exists(filepath):
                msg = f'{filepath} does not exist.'
                raise FileNotFoundError(msg)

            _, ext = os.path.splitext(filepath)
            format_ = ImageFormat.from_extension(ext)

            if format_ is ImageFormat.EXR:
                data, metadata = exrtools.read_exr(filepath)
            else:
                data = np.asarray(pil.open(filepath))

        else:
            msg = f'Object of type {filepath.__class__.__name__} '
            msg += 'is not a str or Path.'
            raise TypeError(msg)

        return Image(data, metadata, format_, allow=True)

    def __init__(self, data, metadata={}, format_=None, allow=False):
        # type: (NDArray, dict[str, Any], Optional[ImageFormat], bool) -> None
        '''
        This constructor should not be called directly except internally and in
        testing.

        Args:
            data (numpy.NDArray): Image.
            metadata (dict, optional): Image metadata. Default: {}.
            format_ (ImageFormat, optional): Format of image. Default: None.
            allow (bool, optional): Whether to allow construction using init.
                Default: False.

        Raises:
            AttributeError: If image data dimensions are not 2 or 3.

        Returns:
            Image: Image instance.
        '''
        if not allow:
            msg = "Please call one of Image's static constructors to create an "
            msg += 'instance. Options include: read, from_array.'
            raise NotImplementedError(msg)

        # ensure data has 3 dimensions
        shape = data.shape
        dims = len(shape)
        if dims > 3 or dims < 2:
            msg = f'Illegal number of dimensions for image data. {dims} not in '
            msg += '[2, 3].'
            raise AttributeError(msg)

        if dims == 2:
            data = data.reshape((*shape, 1))

        self._data = data
        self.metadata = metadata
        self.format = format_
    # --------------------------------------------------------------------------

    def _repr(self):
        # type: () -> str
        fmat = str(None)
        if self.format is not None:
            fmat = self.format.name

        return f'''
       WIDTH: {self.width}
      HEIGHT: {self.height}
NUM_CHANNELS: {self.num_channels}
   BIT_DEPTH: {self.bit_depth.name}
      FORMAT: {fmat}'''[1:]

    def _repr_html_(self):
        # type: () -> None
        '''
        Creates a HTML representation of image data.
        '''
        ImageViewer(self).show()

    def _repr_png(self):
        # type: () -> Optional[bytes]
        '''
        Creates a PNG representation of image data.

        Returns:
            str: PNG.
        '''
        this = self
        if _has_super_brights(self) or _has_super_darks(self):
            this = self.to_unit_space()
        output = this.to_bit_depth(BitDepth.UINT8).to_pil()._repr_png_()
        return output

    def _string_to_channels(self, string):
        # type: (str) -> list
        '''
        Converts string to list of channels.

        Args:
            string (str): String representation of channels.

        Returns:
            list: List of channels.
        '''
        # special rgba short circuit
        combos = [list(combinations('rgba', i)) for i in range(1, 5)]  # type: Any
        combos = list(map(set, chain(*combos)))
        if set(string) in combos:
            return list(string)

        # if channels is actually a layer name
        if string in self.channel_layers:
            found = list(filter(
                lambda x: re.search(string + r'\..+', str(x)),
                self.channels
            ))
            if found != []:
                # found channels that matched [layer-name].[channel] pattern
                return found
            return list(string)
        return [string]

    def __getitem__(self, indices):
        # type: (Union[int, tuple, list, slice, str]) -> Image
        '''
        Gets slice of image data. Indices are given in the order:
        width, height, channel.

        Args:
            indices (int, tuple, list, slice, str): Slice of image data.

        Raises:
            IndexError: If number of indices provided is greater than 3.
            IndexError: If channel given is illegal.
            IndexError: If three lists are given as indices.

        Returns:
            Image: Image slice.
        '''
        if not isinstance(indices, tuple) or isinstance(indices, list):
            indices = [indices]

        size = len(indices)
        if size > 3:
            msg = f'Number of dimensions provided: {size}, is greater than 3.'
            raise IndexError(msg)

        # convert indices to triplet of slices
        columns = slice(None, None)  # type: Any
        rows = slice(None, None)  # type: Any
        channels = slice(None, None)  # type: Any
        if size == 3:
            columns, rows, channels = indices
        elif size == 2:
            columns, rows = indices
        else:
            columns = indices[0]

        # convert channels to list of indices
        channel_meta = self.metadata.get('channels', [])
        if channels.__class__.__name__ in ['str', 'tuple', 'list']:
            if isinstance(channels, str):
                channels = self._string_to_channels(channels)
            channel_meta = channels
            chans = []
            for channel in channels:
                if isinstance(channel, str):
                    if channel not in self.channels:
                        msg = f'{channel} is not a legal channel name.'
                        raise IndexError(msg)
                    channel = self.channels.index(channel)
                chans.append(channel)

            if len(chans) == 1:
                chans = chans[0]
            channels = chans

        # coerce to list for simpler logic
        if isinstance(columns, tuple):
            columns = list(columns)
        if isinstance(rows, tuple):
            rows = list(rows)

        types = [
            columns.__class__.__name__,
            rows.__class__.__name__,
            channels.__class__.__name__,
        ]
        if types == ['list', 'list', 'list']:
            msg = 'Three lists are not acceptable as indices.'
            raise IndexError(msg)

        if isinstance(channels, slice):
            channel_meta = self.channels[channels]

        data = self._data[rows, columns, channels]
        metadata = deepcopy(self.metadata)
        metadata['channels'] = channel_meta
        return Image(data, metadata=metadata, format_=self.format, allow=True)
    # --------------------------------------------------------------------------

    def set_channels(self, channels):
        # type: (list[Union[str, int]]) -> Image
        '''
        Set's channels names.

        Args:
            channels (list[str or int]): List of channel names:

        Raises:
            ValueError: If number of channels given doesn't not equal data
                shape.
            ValueError: If duplicate channel names found.

        Returns:
            Image: self.
        '''
        if len(channels) != self.num_channels:
            msg = 'Number of channels given does not equal last dimension size.'
            msg += f' {len(channels)} != {self.num_channels}.'
            raise ValueError(msg)

        uniq = set(channels)
        if len(uniq) < len(channels):
            for c in uniq:
                channels.remove(c)
            msg = f'Duplicate channel names found: {channels}.'
            raise ValueError(msg)

        metadata = deepcopy(self.metadata)
        metadata['channels'] = channels
        return Image(
            self._data.copy(),
            metadata=metadata,
            format_=self.format,
            allow=True,
        )

    def write(self, filepath, codec=ImageCodec.PIZ):
        # type: (Filepath, ImageCodec) -> None
        '''
        Write image to file.

        Args:
            filepath (str or Path): Full path to image file.
            codec (ImageCodec, optional): EXR compression scheme to be used.
                Default: ImageCodec.PIZ.

        Raises:
            TypeError: If format does not support instance bit depth.
            AttributeError: If format does not support the number of channels in
                instance.
        '''
        if isinstance(filepath, Path):
            filepath = filepath.absolute().as_posix()

        _, ext_ = os.path.splitext(filepath)
        ext = ImageFormat.from_extension(ext_)

        # ensure format is compatible with image data
        if self.bit_depth not in ext.bit_depths:
            msg = f'{ext.name} cannot be written with {self.bit_depth.name}'
            msg += ' data.'
            raise TypeError(msg)

        if self.num_channels > ext.max_channels:
            msg = f'{ext.name} cannot be written with {self.num_channels} '
            msg += f'channels. Max channels supported: {ext.max_channels}.'
            raise AttributeError(msg)

        # write data
        if ext is ImageFormat.EXR:
            metadata = self.metadata
            metadata['channels'] = self.channels
            exrtools.write_exr(filepath, self._data, metadata, codec)

        else:
            pil.fromarray(self._data).save(filepath, format=ext.name)

    def to_bit_depth(self, bit_depth):
        # type: (BitDepth) -> Image
        '''
        Convert image to given bit depth.
        Warning: Numpy's conversions for INT8 are bizarre.

        Args:
            bit_depth (BitDepth): Target bit depth.

        Raises:
            ValueError: If converting from float to 8-bit and values exceed 1.
            ValueError: If converting from float to 8-bit and values less than 0.

        Returns:
            Image: New Image instance at given bit depth.
        '''
        image = self._data
        src = self.bit_depth
        tgt = bit_depth

        if src is tgt:
            return self

        elif src is BitDepth.UINT8 and tgt.type_ is float:
            image = image.astype(tgt.dtype) / 255

        elif src.type_ is float and tgt.bits == 8:
            if _has_super_darks(self):
                msg = f'Image has values below 0. Min value: {image.min()}'
                raise ValueError(msg)

            if _has_super_brights(self):
                msg = f'Image has values above 1. Max value: {image.max()}'
                raise ValueError(msg)

            image = (image * 255).astype(tgt.dtype)

        else:
            image = image.astype(tgt.dtype)

        metadata = deepcopy(self.metadata)
        return Image(image, metadata=metadata, format_=self.format, allow=True)

    def to_unit_space(self):
        # type: () -> Image
        '''
        Normalizes image to [0, 1] range.

        Returns:
            Image: Normalized image.
        '''
        data = self.to_bit_depth(BitDepth.FLOAT32)._data.copy()
        max_, min_ = data.max(), data.min()
        data = (data - min_) / (max_ - min_)
        data = Image.from_array(data).to_bit_depth(self.bit_depth)._data
        metadata = deepcopy(self.metadata)
        return Image(data, metadata=metadata, format_=self.format, allow=True)

    def to_array(self):
        # type: () -> NDArray
        '''
        Returns numpy array.

        Returns:
            numpy.NDArray: Image as numpy array.
        '''
        return self.data

    def to_pil(self):
        # type: () -> pil.Image
        '''
        Returns pil.Image.

        Returns:
            pil: Image as pil.Image.
        '''
        if self.num_channels == 1:
            mode = 'L'
        elif self.num_channels == 3:
            mode = 'RGB'
        elif self.num_channels == 4:
            mode = 'RGBA'
        else:
            raise ValueError('PIL only accepts image with 1, 3 or 4 channels.')
        return pil.fromarray(self.data, mode=mode)

    def compare(self, image, content=False, diff_only=False):
        # type: (Image, bool, bool) -> dict[str, Any]
        '''
        Compare this image with a given image.

        Args:
            image (Image): Image to compare.
            content (bool, optional): If True, compare data. Default: False.
            diff_only (bool, optional): If True, only return the keys with
                differing values. Default: False.

        Raises:
            EnforceError: If image is not an Image instance.
            ValueError: IF content is True and images cannot be compared.

        Returns:
            dict: A dictionary of comparisons.
        '''
        msg = 'Image must be an instance of Image.'
        Enforce(image, 'instance of', Image, message=msg)
        # ----------------------------------------------------------------------

        a = self.info
        b = image.info
        output = {}  # type: dict[str, Any]
        for k, v in a.items():
            output[k] = (v, b.get(k, None))
        for k, v in b.items():
            if k not in output.keys():
                output[k] = (a.get(k, None), v)

        if diff_only:
            for k, v in list(output.items()):
                if v[0] == v[1]:
                    del output[k]

        if content:
            x = self.to_bit_depth(BitDepth.FLOAT16).data
            y = image.to_bit_depth(BitDepth.FLOAT16).data

            try:
                diff = float(abs(x - y).mean())
            except ValueError as e:
                raise ValueError(f'Cannot compare images: {e}')

            output['mean_content_difference'] = diff
            if diff_only and diff == 0:
                del output['mean_content_difference']

        return output

    def __eq__(self, image):
        # type: (object) -> bool
        '''
        Compare this image with a given image.

        Returns:
            bool: True if images are equal.
        '''
        return self.compare(image, content=True, diff_only=True) == {}  # type: ignore
    # --------------------------------------------------------------------------

    @property
    def data(self):
        # type: () -> NDArray
        '''
        numpy.NDArray: Image data.
        '''
        if self.num_channels == 1:
            return np.squeeze(self._data, axis=2)
        return self._data

    @property
    def info(self):
        # type: () -> dict[str, Any]
        '''
        dict: A dictionary of all information about the Image instance.
        '''
        output = dict(
            width=self.width,
            height=self.height,
            channels=self.channels,
            num_channels=self.num_channels,
            bit_depth=self.bit_depth.name,
            dtype=self.bit_depth.dtype,
            bits=self.bit_depth.bits,
            signed=self.bit_depth.signed,
            type=self.bit_depth.type_,
            format_extension=None,
            format_bit_depths=None,
            format_channels=None,
            format_max_channels=None,
            format_custom_metadata=None,
        )
        if self.format is not None:
            fmat = dict(
                format_extension=self.extension,
                format_bit_depths=self.format.bit_depths,
                format_channels=self.format.channels,
                format_max_channels=self.format.max_channels,
                format_custom_metadata=self.format.custom_metadata,
            )
            output.update(fmat)
        return output

    @property
    def shape(self):
        # type: () -> Tuple[int, int, int]
        '''
        tuple[int]: (width, height, channels) of image.
        '''
        return (self.width, self.height, self.num_channels)

    @property
    def width(self):
        # type: () -> int
        '''
        int: Width of image.
        '''
        return self._data.shape[1]

    @property
    def height(self):
        # type: () -> int
        '''
        int: Height of image.
        '''
        return self._data.shape[0]

    @property
    def width_and_height(self):
        # type: () -> Tuple[int, int]
        '''
        tupe[int]: (width, height) of image.
        '''
        return (self.width, self.height)

    @property
    def channels(self):
        # type: () -> list[Union[str, int]]
        '''
        list[str or int]: List of channel names.
        '''
        if 'channels' in self.metadata:
            return self.metadata['channels']
        return cvt.get_channels_from_array(self._data)

    @property
    def num_channels(self):
        # type: () -> int
        '''
        int: Number of channels in image.
        '''
        return len(self.channels)

    @property
    def max_channels(self):
        # type: () -> Optional[int]
        '''
        int: Maximum number of channels supported by image format.
        '''
        if self.format is None:
            return None
        return self.format.max_channels

    @property
    def channel_layers(self):
        # type: () -> list[str]
        '''
        list[str]: List of channel layers.
        '''
        channels = [str(x) for x in self.channels]
        with_layer = list(filter(lambda x: '.' in str(x), channels))  # type: list[str]
        wo_layer_name = list(filter(lambda x: '.' not in str(x), channels))

        # break out channels without layer names into groups of 4
        len_ = len(wo_layer_name)
        layers = []
        for idx in range(0, len_, 4):
            layer_ = wo_layer_name[idx:min(idx + 4, len_)]
            layer = ''.join(layer_)
            layers.append(layer)

        # append all unique layers of channels with layer names
        for chan in with_layer:
            layer = ''.join(chan.split('.')[0])
            if layer not in layers:
                layers.append(layer)
        return layers

    @property
    def bit_depth(self):
        # type: () -> BitDepth
        '''
        BitDepth: Bit depth of image.
        '''
        return BitDepth.from_dtype(self._data.dtype)

    @property
    def extension(self):
        # type: () -> Optional[str]
        '''
        str: Image format extension.
        '''
        if self.format is None:
            return None
        return self.format.extension



import subprocess

import click
# ------------------------------------------------------------------------------

'''
Command line interface to cv-depot library
'''


@click.group()
def main():
    pass


@main.command()
def bash_completion():
    '''
    BASH completion code to be written to a _cv-depot completion file.
    '''
    cmd = '_cv_depot_COMPLETE=bash_source cv-depot'
    result = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    result.wait()
    click.echo(result.stdout.read())


@main.command()
def zsh_completion():
    '''
    ZSH completion code to be written to a _cv-depot completion file.
    '''
    cmd = '_cv_depot_COMPLETE=zsh_source cv-depot'
    result = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    result.wait()
    click.echo(result.stdout.read())


if __name__ == '__main__':
    main()



from typing import Optional  # noqa F401
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from cv_depot.core.image import Image  # noqa: F401

import ipywidgets as ipy
import IPython.display as ipython
import re

import cv_depot.api as cvd
# ------------------------------------------------------------------------------


class ImageViewer:
    def __init__(self, image, size=87, gamma=1, premultiply=False):
        # type: (Image, int, float, bool) -> None
        '''
        Constructs an ImageViewer widget, used for displaying Image instances.

        Args:
            image (Image): Image instance.
            size (int, optional): Image size in percentage. Default: 87.
            gamma (float, optional): Initial gamma value. Default: 1.
            premultiply (bool, optional): Premultiply image by last channel.
                Default: False.

        Raises:
            EnforceError: If image is not an Image instance.
        '''
        # image
        self._image = image
        self.size = size
        self.gamma = gamma
        self.premult = premultiply

        # layer
        self.layer = self._get_layer_options()[0]
        self.layer_selector = ipy.Dropdown(
            description='layer',
            label=self.layer,
            options=self._get_layer_options(),
        )
        self.layer_selector.observe(self._handle_layer_event, names='value')

        # channel
        self.channel = self._get_channel_options()[0]
        self.channel_selector = ipy.Dropdown(
            description='channel',
            label=self.channel,
            options=self._get_channel_options(),
        )
        self.channel_selector.observe(self._handle_channel_event, names='value')

        # size slider
        self.size_slider = ipy.IntSlider(
            value=self.size, min=0, max=100, step=1, names='value', description='size'
        )
        self.size_slider.observe(self._handle_resize_event, names='value')

        # gamma slider
        self.gamma_slider = ipy.FloatSlider(
            value=self.gamma, min=0, max=10, step=0.01, names='value', description='gamma'
        )
        self.gamma_slider.observe(self._handle_gamma_event, names='value')

        self.premult_checkbox = ipy.Checkbox(
            value=False, description='premultiply',
        )
        self.premult_checkbox.observe(self._handle_premult_event, names='value')

        # viewer
        self.viewer = ipy.Image(value=self._get_png(), width=f'{self.size}%')
        self.info = ipy.HTML(value=self._get_info())
        space = ipy.HTML(value='<div style="width: 115px;"></div>')

        # widgets
        self._widgets = [
            ipy.HBox(
                [
                    space,
                    self.layer_selector,
                    self.channel_selector,
                    self.size_slider,
                    self.gamma_slider,
                    self.premult_checkbox,
                ],
                layout=ipy.Layout(flex_flow='row')
            ),
            ipy.HBox(
                [self.info, self.viewer],
                layout=ipy.Layout(flex_flow='row')
            )
        ]

    def show(self):
        # type: () -> None
        '''
        Call ipython.display with widgets.
        '''
        ipython.display(*self._widgets)

    def _get_layer_options(self):
        # type: () -> list[str]
        '''
        Get list of channel layers.

        Returns:
            list[str]: List of channel layers.
        '''
        return self._image.channel_layers

    def _get_channel_options(self):
        # type: () -> list
        '''
        Get list of channel options.

        Returns:
            list: List of channel options.
        '''
        chan = self._image[:, :, self.layer].channels
        return ['all'] + chan

    def _get_info(self):
        # type: () -> str
        '''
        Creates a HTML representation of image info.

        Returns:
            str: HTML.
        '''
        desc = self._image._repr()
        desc = re.sub('<', '&lt;', desc)
        desc = re.sub('>', '&gt;', desc)
        desc = re.sub('\n', '<br>', desc)
        desc = re.sub(' ', '&nbsp;', desc)
        desc = f'<p style="font-family: monospace; font-size: 13px;">{desc}</p>'
        return desc

    def _get_png(self):
        # type: () -> Optional[bytes]
        '''
        Creates a PNG representation of image data.

        Returns:
            str: PNG.
        '''
        chan = self.channel
        if chan == 'all':
            chan = self.layer
        chans = self._image[:, :, chan].channels
        if not self.premult and len(chans) > 3:
            chans = chans[:3]
        return cvd.ops.filter.gamma(self._image[:, :, chans], self.gamma)._repr_png()

    def _handle_layer_event(self, event):
        # type: (dict) -> None
        '''
        Handles layer selector events.

        Args:
            event (dict): Event.
        '''
        if event['type'] == 'change':
            # layer
            self.layer = event['new']
            self.layer_selector.value = self.layer

            # channel
            options = self._get_channel_options()
            chan = options[0]
            self.channel = chan
            self.channel_selector.options = options
            self.channel_selector.label = chan
            self.channel_selector.value = chan

            # viewer
            self.viewer.value = self._get_png()

    def _handle_channel_event(self, event):
        # type: (dict) -> None
        '''
        Handles channel selector events.

        Args:
            event (dict): Event.
        '''
        if event['type'] == 'change':
            self.channel = event['new']
            self.channel_selector.value = self.channel
            self.viewer.value = self._get_png()

    def _handle_resize_event(self, event):
        # type: (dict) -> None
        '''
        Handles image resize events.

        Args:
            event (dict): Event.
        '''
        if event['type'] == 'change':
            self.size = event['new']
            self.size_slider.value = self.size
            self.viewer.width = f'{self.size}%'

    def _handle_gamma_event(self, event):
        # type: (dict) -> None
        '''
        Handles image resize events.

        Args:
            event (dict): Event.
        '''
        if event['type'] == 'change':
            self.gamma = event['new']
            self.gamma_slider.value = self.gamma
            self.viewer.value = self._get_png()

    def _handle_premult_event(self, event):
        # type: (dict) -> None
        '''
        Handles premultiply events.

        Args:
            event (dict): Event.
        '''
        if event['type'] == 'change':
            self.premult = event['new']
            self.premult_checkbox.value = self.premult
            self.viewer.value = self._get_png()



from typing import Any, Dict, Union  # noqa F401
from cv_depot.core.types import Filepath  # noqa F401

import os
from pathlib import Path
import re

import ffmpeg
from lunchbox.enforce import Enforce

from cv_depot.core.enum import VideoCodec, VideoFormat
# ------------------------------------------------------------------------------


def get_video_metadata(filepath):
    # type: (Filepath) -> Dict[str, Any]
    '''
    Retrieve video file metadata with ffmpeg.

    Args:
        filepath (str or Path): Path to video file.

    Raises:
        EnforceError: If filepath is not a file or does not exist.
        EnforceError: If filepath format is unsupported.

    Returns:
        dict: Metadata.
    '''
    filepath = Path(filepath)

    msg = f'{filepath} is not a file or does not exist.'
    Enforce(filepath.is_file(), '==', True, message=msg)

    ext = filepath.suffix[1:]
    formats = ['m4v', 'mov', 'mp4', 'mpg', 'mpeg']
    msg = 'Illegal format: {a}. Legal formats: {b}.'
    Enforce(ext.lower(), 'in', formats, message=msg)
    # --------------------------------------------------------------------------

    meta = ffmpeg.probe(filepath.as_posix())['streams'][0]
    lut = {'16': 'FLOAT16', '32': 'FLOAT32', '8': 'UINT8'}
    output = dict(
        width=meta['width'],
        height=meta['height'],
        num_frames=int(meta['nb_frames']),
        frame_rate=int(meta['r_frame_rate'].split('/')[0]),
        bit_depth=lut[meta['bits_per_raw_sample']],
        codec=meta['codec_name'],
    )
    return output


def write_video(
    source,                  # type: Filepath
    target,                  # type: Filepath
    framerate=24,            # type: int
    codec=VideoCodec.H264,   # type: VideoCodec
    format_=VideoFormat.MP4  # type: VideoFormat
):
    # type: (...) -> None
    '''
    Writes given input file or file pattern to given target file.

    Args:
        source (str or Path): Source image filepath or file pattern.
        target (str or Path): Target file.
        framerate (int, optional): Video framerate. Default: 24 fps.
        codec (VideoCodec, optional): Video codec. Default: VideoCodec.H264.
        format_ (VideoFormat, optional): Video container format.
            Default: VideoFormat.MP4

    Raises:
        EnforceError: If source is not a filepath or file pattern.
        EnforceError: framerate is not an integer greater than 0.
        EnforceError: If codec is illegal.
        EnforceError: If format is illegal.
    '''
    source = Path(source).as_posix()
    target = Path(target).as_posix()
    if not Path(source).is_file():
        msg = 'Source must be a video file or file pattern with {a} in it. '
        msg += 'Given value: {b}.'
        Enforce('{frame}', 'in', source, message=msg)
        source = re.sub(r'\{frame\}', '%04d', source)

    msg = 'Framerate must be an integer greater than 0. Given value: {a}.'
    Enforce(framerate, 'instance of', int, message=msg)
    Enforce(framerate, '>', 0, message=msg)

    Enforce(codec, 'instance of', VideoCodec)
    Enforce(format_, 'instance of', VideoFormat)
    # --------------------------------------------------------------------------

    os.makedirs(Path(target).parent, exist_ok=True)
    ffmpeg \
        .input(source, framerate=framerate) \
        .output(target, vcodec=codec.ffmpeg_code, format=format_.extension) \
        .run(overwrite_output=True)



from typing import Union  # noqa F401

import re

from lunchbox.enforce import Enforce
import cv2
import numpy as np

from cv_depot.core.channel_map import ChannelMap
from cv_depot.core.color import BasicColor
from cv_depot.core.enum import BitDepth
from cv_depot.core.image import Image
import cv_depot.core.image as cvimg
import cv_depot.ops.draw as cvdraw
# ------------------------------------------------------------------------------


def has_super_brights(image):
    # type: (Image) -> bool
    '''
    Determines if given image has values above 1.0.

    Args:
        image (Image): Image instance.

    Raises:
        EnforceError: If image is not an Image instance.

    Returns:
        bool: Presence of super brights.
    '''
    return cvimg._has_super_brights(image)


def has_super_darks(image):
    # type: (Image) -> bool
    '''
    Determines if given image has values below 0.0.

    Args:
        image (Image): Image instance.

    Raises:
        EnforceError: If image is not an Image instance.

    Returns:
        bool: Presence of super darks.
    '''
    return cvimg._has_super_darks(image)


def to_hsv(image):
    # type: (Image) -> Image
    '''
    Convert image to hue, saturation, value colorspace.

    Args:
        Image: Image to be converted.

    Raises:
        AttributeError: If given image does not have RGB channels.

    Returns:
        Image: Image converted to HSV.
    '''
    rgb = list('rgb')
    channels = set(image.channels).intersection(rgb)
    if channels != set(rgb):
        msg = 'Image does not contain RGB channels. '
        msg += f'Channels found: {image.channels}.'
        raise AttributeError(msg)

    img_ = image.to_bit_depth(BitDepth.FLOAT32)
    img = img_[:, :, rgb].data
    img = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
    img[:, :, 0] = np.divide(img[:, :, 0], 360)
    output = Image\
        .from_array(img)\
        .set_channels(list('hsv'))\
        .to_bit_depth(image.bit_depth)
    return output


def to_rgb(image):
    # type: (Image) -> Image
    '''
    Convert image from HSV to RGB.

    Args:
        Image: Image to be converted.

    Raises:
        AttributeError: If given image does not have RGB channels.

    Returns:
        Image: Image converted to RGB.
    '''
    hsv = list('hsv')
    channels = set(image.channels).intersection(hsv)
    if channels != set(hsv):
        msg = 'Image does not contain HSV channels. '
        msg += f'Channels found: {image.channels}.'
        raise AttributeError(msg)

    img_ = image.to_bit_depth(BitDepth.FLOAT32)
    img = img_[:, :, hsv].data
    img = cv2.cvtColor(img, cv2.COLOR_HSV2RGB)
    img[:, :, 0] = np.divide(img[:, :, 0], 360)
    output = Image\
        .from_array(img)\
        .set_channels(list('rgb'))\
        .to_bit_depth(image.bit_depth)
    return output


def invert(image):
    # type: (Image) -> Image
    '''
    Inverts the values of the given image.
    Black becomes white, white becomes black.

    Args:
        image (Image): Image to be inverted.

    Raises:
        EnforeError: If image is not an instance of Image.

    Returns:
        image: Image
    '''
    Enforce(image, 'instance of', Image)
    # --------------------------------------------------------------------------

    bit_depth = image.bit_depth
    data = image.to_bit_depth(BitDepth.FLOAT32).data
    data = data * -1 + 1
    output = Image.from_array(data).to_bit_depth(bit_depth)
    return output


def mix(a, b, amount=0.5):
    # type: (Image, Image, float) -> Image
    '''
    Mix images A and B by a given amount.
    An amount of 1.0 means 100% of image A.
    An amount of 0.0 means 100% of image B.

    Args:
        a (Image): Image A.
        b (Image): Image B.
        amount (float, optional): Amount of image A. Default: 0.5

    Raises:
        EnforceError: If a is not an Image instance.
        EnforceError: If b is not an Image instance.
        EnforceError: If amount is not between 0 and 1.

    Returns:
        Image: Mixture of A and B.
    '''
    Enforce(a, 'instance of', Image)
    Enforce(b, 'instance of', Image)
    Enforce(amount, '<=', 1)
    Enforce(amount, '>=', 0)
    # --------------------------------------------------------------------------

    amount = float(amount)
    bit_depth = a.bit_depth
    x = a.to_bit_depth(BitDepth.FLOAT32).data
    y = b.to_bit_depth(BitDepth.FLOAT32).data
    img = x * amount + y * (1 - amount)
    output = Image.from_array(img).to_bit_depth(bit_depth)
    return output


def remap_single_channel(image, channels):
    # type: (Image, list) -> Image
    '''
    Maps an image with a single channel to an image of a given number of
    channels.

    Args:
        image (Image): Image to be mapped.
        channels (list): List of channel names to map image to.

    Raises:
        EnforceError: If image is not an Image with only one channel.
        EnforceError: If channels is not a list.

    Returns:
        Image: Image with given channels.
    '''
    Enforce(image, 'instance of', Image)
    msg = 'Image must be an Image with only 1 channel. '
    msg += 'Channels found: {a}.'
    Enforce(image.num_channels, '==', 1, message=msg)
    Enforce(channels, 'instance of', list)
    # --------------------------------------------------------------------------

    data = np.squeeze(image.data)[..., np.newaxis]
    output = np.concatenate([data] * len(channels), axis=2)
    output = Image.from_array(output).set_channels(channels)
    return output


def remap(images, channel_map):
    # type: (Union[Image, list[Image]], ChannelMap) -> Image
    '''
    Maps many images into a single image according to a given channel
    map.

    Args:
        images (Image, list[Image]): Images.
        channel_map (ChannelMap): Mapping of image channels into output image.

    Raises:
        EnforceError: If images is not an instance of Image or list of Images.
        EnforceError: If images are not of all the same width and height.
        EnforceError: If channel_map is not an instance of ChannelMap.

    Returns:
        Image: Combined image.
    '''
    if isinstance(images, Image):
        images = [images]

    msg = 'Images must be an Image or list of Images of uniform width and height.'
    msg += f' Given type: {images.__class__.__name__}'
    [Enforce(x, 'instance of', Image) for x in images]

    shapes = {x.width_and_height for x in images}
    Enforce(len(shapes), '==', 1, message=msg)

    Enforce(channel_map, 'instance of', ChannelMap)
    # --------------------------------------------------------------------

    w, h = images[0].width_and_height
    bit_depth = images[0].bit_depth
    images = [x.to_bit_depth(bit_depth) for x in images]

    shape = (w, h, 1)
    black = cvdraw.swatch(shape, BasicColor.BLACK, bit_depth=bit_depth)
    white = cvdraw.swatch(shape, BasicColor.WHITE, bit_depth=bit_depth)

    channels = []
    for chan in channel_map.source:
        chan_l = chan.lower()
        img = None

        if chan_l in ['b', 'black']:
            img = black
        elif chan_l in ['w', 'white']:
            img = white
        else:
            frame, tgt_chan = re.split(r'\.', chan, maxsplit=1)
            img = images[int(frame)][:, :, tgt_chan]

        array = np.squeeze(img.data)[..., np.newaxis]
        channels.append(array)

    img = np.concatenate(channels, axis=2).astype(bit_depth.dtype)
    output = Image.from_array(img).set_channels(channel_map.target)  # type: ignore
    return output



'''
This moudle is meant as a convenience for programmers who want a clean namespace
to sift through.
'''

import argparse as __argparse
import inspect as __inspect
import re as __re
import types as __types

import cv_depot.ops as __ops  # noqa F401
from cv_depot.core.channel_map import ChannelMap  # noqa F401
from cv_depot.core.color import BasicColor, Color  # noqa F401
from cv_depot.core.enum import BitDepth, ImageFormat, VideoCodec, VideoFormat  # noqa F401
from cv_depot.core.enum import Anchor, Direction  # noqa F401
from cv_depot.core.image import Image, ImageCodec  # noqa F401
from cv_depot.core.viewer import ImageViewer  # noqa F401
# ------------------------------------------------------------------------------


def __create_namespace(module):
    # type: (object) -> __argparse.Namespace
    '''
    Creates a clean namespace from a module.
    Only grabs public functions from the module.

    Args:
        module (object): module

    Returns:
        argparse.Namespace: Clean namespace.
    '''
    params = {}
    for key, val in __inspect.getmembers(module):
        if isinstance(val, __types.FunctionType):
            if __re.search('^[a-z]', key):
                params[key] = val
    return __argparse.Namespace(**params)


ops = __argparse.Namespace(
    channel=__create_namespace(__ops.channel),
    draw=__create_namespace(__ops.draw),
    edit=__create_namespace(__ops.edit),
    filter=__create_namespace(__ops.filter),
)



from typing import Union  # noqa F401
from numpy.typing import NDArray  # noqa F401
# ------------------------------------------------------------------------------


def get_channels_from_array(array):
    # type: (NDArray) -> list[Union[str, int]]
    '''
    Returns a list of strings representing the given array's channels.
    If array has only one channel then ['l'] is returned.
    First 4 channels are [r, g, b, a], in that order. All subsequent channels
    are integers starting at 4.

    Args:
        array (numpy.NDArray): Numpy array with 2+ dimensional shape.

    Returns:
        list[str and int]: Channels.
    '''
    if len(array.shape) < 3 or array.shape[2] == 1:
        return ['l']
    else:
        temp = list(range(array.shape[2]))
        lut = {0: 'r', 1: 'g', 2: 'b', 3: 'a'}  # type: dict[int, str]
        channels = []  # type: list[Union[str, int]]
        for i in temp:
            c = i  # type: Union[int, str]
            if i in lut:
                c = lut[i]
            channels.append(c)
        return channels



from typing import Iterable  # noqa F401

from lunchbox.enforce import Enforce
# ------------------------------------------------------------------------------


def enforce_homogenous_type(iterable, name='Iterable'):
    # type: (Iterable, str) -> None
    '''
    Ensures that iterable only contains only one type of object.

    Args:
        items (iterable): Iterable.
        name (str, optional): First word in error message. Default: Iterable.

    Raises:
        EnforceError: If iterable contains more than one type of object.
    '''
    types = [x.__class__.__name__ for x in iterable]
    types = sorted(list(set(types)))
    msg = f'{name} may only contain one type of object. Found types: {types}.'
    Enforce(len(types), '==', 1, message=msg)



from collections import OrderedDict
import re

from lunchbox.enforce import Enforce
# ------------------------------------------------------------------------------


class ChannelMap(OrderedDict):
    '''
    A channel to channel mapping between multiple source images and one target
    image.
    '''
    def __repr__(self):
        # type () -> str
        '''
        Returns a table of with columns: order, target and source.
        '''
        max_i = max(len(self.items()) - 1, len('order'))
        max_k = max(max(map(len, self.keys())), len('target'))
        header = f'{{i:<{max_i}}} {{k:<{max_k}}}       {{v}}'
        header = header.format(i='order', k='target', v='source')
        output = [header, '-' * len(header)]

        pattern = f'{{i:<{max_i}}} {{k:<{max_k}}}  <--  {{v}}'
        for i, (k, v) in enumerate(self.items()):
            output.append(pattern.format(i=i, k=k, v=v))
        output = '\n'.join(output)
        return output

    def __setitem__(self, key, value):
        # type: (str, str) -> None
        '''
        Sets given key to given value.
        Ensures that given key and value are legal.

        Args:
            key (str): Key.
            value (str): Value.

        Raises:
            EnforceError: If key is not a string.
            EnforceError: If value is not a string.
            EnforceError: If value is not one of: b, w, black, white,
                [frame].[channel]
        '''
        Enforce(key, 'instance of', str)
        Enforce(value, 'instance of', str)
        result = re.match(r'^(b|w|black|white|\d+\..+)$', value.lower())
        msg = 'ChannelMap values must match: b, w, black, white, or '
        msg += fr'[frame].[channel]. Illegal value: {value}.'
        Enforce(result, '!=', None, message=msg)

        super().__setitem__(key, value)

    @property
    def source(self):
        # type: () -> list[str]
        '''
        list[str]: Ordered list of source channels.
        '''
        return list(self.values())

    @property
    def target(self):
        # type: () -> list[str]
        '''
        list[str]: Ordered list of target channels.
        '''
        return list(self.keys())



from typing import Any, List, Tuple, Union  # noqa F401
from cv_depot.core.types import AnyColor  # noqa F401

import copy

from lunchbox.enforce import Enforce
import cv2
import numpy as np

from cv_depot.core.channel_map import ChannelMap
from cv_depot.core.color import BasicColor, Color
from cv_depot.core.image import BitDepth, Image
import cv_depot.ops.channel as cvchan
import cv_depot.ops.draw as cvdraw
import cv_depot.ops.filter as cvfilt
# ------------------------------------------------------------------------------


def swatch(
    shape,                      # type: Union[Tuple[int, int, int], List[int]]
    color,                      # type: AnyColor
    fill_value=0.0,             # type: float
    bit_depth=BitDepth.FLOAT32  # type: BitDepth
):
    # type: (...) -> Image
    '''
    Creates an image of the given shape and color.

    Args:
        shape (tuple[int]): List of 3 ints.
        color (Color): Color of swatch RGB or L channels.
        fill_value (float, optional): Value to fill additional channels with.
            Default: 0.
        bit_depth (BitDepth): Bit depth of swatch. Default: BitDepth.FLOAT32

    Raises:
        EnforceError: If shape is not a tuple of 3 integers.
        EnforceError: If shape has any zero dimensions.

    Returns:
        Image: Color swatch.
    '''
    msg = f'Shape must be an tuple or list of 3 integers. Given shape: {shape}. '
    msg += f'Given type: {type(shape)}.'
    Enforce(shape.__class__.__name__, 'in', ['tuple', 'list'], message=msg)
    Enforce(len(shape), '==', 3, message=msg)
    for item in shape:
        Enforce(item, 'instance of', int, message=msg)

    msg = 'Illegal shape: Each shape dimension must be greater than 0. '
    msg += f'Given shape: {shape}.'
    Enforce(min(shape), '>', 0, message=msg)
    # --------------------------------------------------------------------------

    w, h, c = shape

    if isinstance(color, str):
        color = BasicColor.from_string(color)
    if isinstance(color, BasicColor):
        color = Color.from_basic_color(color)
    color = Color.from_array(
        color.to_array(), num_channels=c, fill_value=fill_value
    )

    output = np.ones((h, w, c), np.float32) * color.to_array()
    output = Image.from_array(output).to_bit_depth(bit_depth)
    return output


def grid(image, shape, color=BasicColor.CYAN.name, thickness=1):
    # type: (Image, Tuple[int, int], AnyColor, int) -> Image
    '''
    Draws a grid on a given image.

    Args:
        image (Image): Image to be drawn on.
        shape (tuple[int]): Width, height tuple.
        color (Color, optional): Color of grid. Default: BasicColor.CYAN.
        thickness (int, optional): Thickness of grid lines in pixels.
            Default: 1.

    Raises:
        EnforceError: If image is not an Image instance.
        EnforceError: If shape is not of length 2.
        EnforceError: If width or height of shape is less than 0:
        EnforceError: If color is not an instance of Color or BasicColor.
        EnforceError: If thickness is not greater than 0.

    Returns:
        Image: Image with grid.
    '''
    msg = 'Illegal image. {a.__class__.__name__} is not an instance of Image.'
    Enforce(image, 'instance of', Image, message=msg)
    msg = f'Illegal shape. Expected (w, h). Found: {shape}.'
    Enforce(len(shape), '==', 2, message=msg)
    msg = 'Shape width must be greater than or equal to 0. {a} < 0.'
    Enforce(shape[0], '>=', 0, message=msg)
    msg = 'Shape height must be greater than or equal to 0. {a} < 0.'
    Enforce(shape[1], '>=', 0, message=msg)
    msg = 'Color type must be in {b}. Found: {a}.'
    Enforce(color.__class__.__name__, 'in', ['Color', 'BasicColor'], message=msg)
    msg = 'Line thickness must be an integer. Found: {a.__class__.__name__}.'
    Enforce(thickness, 'instance of', int, message=msg)
    msg = 'Line thickness must be greater than 0. {a} !> 0.'
    Enforce(thickness, '>', 0)
    # --------------------------------------------------------------------------

    w0, h0 = shape
    w, h = image.width_and_height
    bit_depth = image.bit_depth
    img = image.to_bit_depth(BitDepth.FLOAT32).data.copy()
    if isinstance(color, str):
        color = BasicColor.from_string(color)
    if isinstance(color, BasicColor):
        color = Color.from_basic_color(color)
    clr = color.to_array().tolist()  # type: Any

    # vertical lines
    w_step = int(round(w / w0, 0))
    for x in range(w_step, w, w_step):
        img = cv2.line(img, (x, 0), (x, h), clr, thickness)

    # horizontal lines
    h_step = int(round(h / h0, 0))
    for y in range(h_step, h, h_step):
        img = cv2.line(img, (0, y), (w, y), clr, thickness)

    output = Image.from_array(img).to_bit_depth(bit_depth)
    return output


def checkerboard(tiles_wide, tiles_high, tile_shape=(10, 10)):
    # type: (int, int, Tuple[int, int]) -> Image
    '''
    Draws a checkerboard of given width, height and tile shape.

    Args:
        tiles_wide (int): Number of tiles wide checkerboard will be.
        tiles_high (int): Number of tiles high checkerboard will be.
        tile_shape (tuple[int], optional): Width, height tuple of tile shape.
            Default: (10, 10).

    Raises:
        EnforceError: If tiles_wide is not greater than 0.
        EnforceError: If tiles_high is not greater than 0.
        EnforceError: If tile width is not greater than 0.
        EnforceError: If tile height is not greater than 0.

    Returns:
        Image: Checkerboard image.
    '''
    w, h = tile_shape
    shape = (w, h, 3)

    msg = 'Tiles_wide must be greater than 0. {a} !> 0.'
    Enforce(tiles_wide, '>', 0, message=msg)
    msg = 'Tiles_high must be greater than 0. {a} !> 0.'
    Enforce(tiles_high, '>', 0, message=msg)
    msg = 'Tile width must be greater than 0. {a} !> 0.'
    Enforce(w, '>', 0, message=msg)
    msg = 'Tile height must be greater than 0. {a} !> 0.'
    Enforce(h, '>', 0, message=msg)
    # --------------------------------------------------------------------------

    black = swatch(shape, BasicColor.BLACK).data
    white = swatch(shape, BasicColor.WHITE).data
    even = []
    odd = []
    for i in range(0, tiles_wide):
        if i == 0 or i % 2 == 0:
            even.append(black)
            odd.append(white)
        else:
            even.append(white)
            odd.append(black)

    even = np.concatenate(even, axis=1)
    odd = np.concatenate(odd, axis=1)

    rows = []
    for i in range(0, tiles_high):
        if i == 0 or i % 2 == 0:
            rows.append(even)
        else:
            rows.append(odd)

    output = Image.from_array(np.concatenate(rows, axis=0))
    return output


def highlight(
    image, mask='a', opacity=0.5, color=BasicColor.CYAN2.name, inverse=False
):
    # type: (Image, str, float, AnyColor, bool) -> Image
    '''
    Highlight a masked portion of a given image according to a given channel.

    Args:
        image (Image): Image to be highlighted.
        mask (str, optional): Channel to be used as mask. Default: alpha.
        opacity (float, optional): Opacity of highlight overlayed on image.
            Default: 0.5
        color (Color or BasicColor, optional): Color of highlight.
            Default: BasicColor.CYAN2.
        inverse (bool, optional): Whether to invert the highlight.
            Default: False.

    Raises:
        EnforceError: If image is not an instance of Image.
        EnforceError: If mask is not an instance of str.
        EnforceError: If mask not found in image channels.
        EnforceError: If opacity is < 0 or > 1.
        EnforceError: If color is not instance of Color.
        EnforceError: If inverse is not a boolean.

    Returns:
        Image: Highlighted image.
    '''
    if isinstance(color, str):
        color = BasicColor.from_string(color)
    if isinstance(color, BasicColor):
        color = Color.from_basic_color(color)

    Enforce(image, 'instance of', Image)
    Enforce(mask, 'instance of', str)
    msg = 'Mask channel: {a} not found in image channels: {b}.'
    Enforce(mask, 'in', image.channels, message=msg)
    Enforce(opacity, '>=', 0)
    Enforce(opacity, '<=', 1)
    Enforce(color, 'instance of', Color)
    Enforce(inverse, 'instance of', bool)
    # --------------------------------------------------------------------------

    img = image.to_bit_depth(BitDepth.FLOAT32)
    channels = copy.deepcopy(image.channels)
    matte = cvchan.remap_single_channel(img[:, :, mask], channels)
    imatte = cvchan.invert(matte)
    if inverse:
        matte, imatte = imatte, matte

    swatch = cvdraw.swatch(image.shape, color, fill_value=1.0)
    data = (image.data * imatte.data) + (swatch.data * matte.data)
    output = Image.from_array(data).to_bit_depth(image.bit_depth)
    output = cvchan \
        .mix(image, output, amount=1 - opacity) \
        .set_channels(image.channels)
    return output


def outline(image, mask='a', width=10, color=BasicColor.CYAN2.name):
    # type (Image, str, int, AnyColor) -> Image
    '''
    Use a given mask to outline a given image.

    Args:
        image (Image): Image with mask channel.
        mask (str, optional): Mask channel. Default: alpha.
        width (int, optional): Outline width. Default: 10.
        color (Color or BasicColor, optional): Color of outline.
            Default: BasicColor.CYAN2

    Raises:
        EnforceError: If image is not an instance of Image.
        EnforceError: If channel not found in image channels.
        EnforceError: If width is not >= 0.
        EnforceError: If color is not an instance of Color or BasicColor.

    Returns:
        Image: Image with outline.
    '''
    Enforce(image, 'instance of', Image)
    msg = 'Mask channel: {a} not found in image channels: {b}.'
    Enforce(mask, 'in', image.channels, message=msg)
    Enforce(width, '>=', 0)
    # --------------------------------------------------------------------------

    cmap = ChannelMap({c: f'0.{c}' for c in image.channels})
    cmap[mask] = '1.l'
    w = int(round(width / 2, 0))
    edge = cvfilt.canny_edges(image[:, :, mask], size=w)
    output = cvchan.remap([image, edge], cmap)
    output = highlight(output, mask=mask, opacity=1.0, color=color)
    return output


def annotate(
    image,                        # type: Image
    mask='a',                     # type: str
    opacity=0.5,                  # type: float
    width=10,                     # type: int
    color=BasicColor.CYAN2.name,  # type: AnyColor
    inverse=False,                # type: bool
    keep_mask=False,              # type: bool
):
    # type (...) -> Image
    '''
    Annotate a given image according to a given mask channel.

    Args:
        image (Image): Image with mask channel.
        mask (str, optional): Mask channel. Default: alpha.
        opacity (float, optional): Opacity of annotation. Default: 0.5
        width (int, optional): Outline width. Default: 10.
        color (Color or BasicColor, optional): Color of outline.
            Default: BasicColor.CYAN2
        inverse (bool, optional): Whether to invert the annotation.
            Default: False.
        keep_mask (bool, optional): Whether to keep the mask channel.
            Default: False.

    Returns:
        Image: Image with outline.
    '''
    output = highlight(
        image, mask=mask, opacity=opacity, color=color, inverse=inverse
    )
    output = outline(output, mask=mask, width=width, color=color)
    if not keep_mask:
        channels = copy.deepcopy(image.channels)
        channels.remove(mask)
        output = output[:, :, channels]
    return output
